/*
 * This file exports the experimental dtype API as exposed via the
 * `numpy/core/include/numpy/experimental_dtype_api.h`
 * header file.
 *
 * This file is a stub, all important definitions are in the code file.
 *
 * NOTE: This file is considered in-flux, exploratory and transitional.
 */

#ifndef NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_
#define NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_

NPY_NO_EXPORT PyObject *
_get_experimental_dtype_api(PyObject *mod, PyObject *arg);


#endif  /* NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_ */
